import re
import nest_asyncio  # type: ignore # necessary import fix required for async to work in notebooks

from biolib import cli_utils

nest_asyncio.apply()


class CompletedProcess:
    def __init__(self, stdout, stderr, exitcode):
        self.stdout = stdout
        self.stderr = stderr
        self.exitcode = exitcode

    def __str__(self):
        return cli_utils.get_pretty_print_module_output_string(self.stdout, self.stderr, self.exitcode)

    def ipython_markdown(self):
        import IPython.display  # type:ignore # pylint: disable=import-error, import-outside-toplevel
        markdown_str = self.stdout.decode('utf-8')
        # prepend ./biolib_results/ to all paths
        # ie [SeqLogo](./SeqLogo2.png) test ![SeqLogo](./SeqLogo.png)
        # ![SeqLogo](SeqLogo.png)  ![SeqLogo](/SeqLogo.png)
        # is transformed to ![SeqLogo](./biolib_results/SeqLogo2.png) test ![SeqLogo](./biolib_results/SeqLogo.png)
        # ![SeqLogo](./biolib_results/SeqLogo.png)  ![SeqLogo](./biolib_results/SeqLogo.png)
        markdown_str_modified = re.sub(
            r'\!\[([^\]]*)\]\((\.\/|\/|)([^\)]*)\)',
            r'![\1](./biolib_results/\3)',
            markdown_str,
        )
        IPython.display.display(IPython.display.Markdown(markdown_str_modified))



from biolib.typing_utils import TypedDict


class VsockProxyResponse(TypedDict):
    hostname: str
    id: str
    port: int


class ComputeNodeInfo(TypedDict):
    auth_token: str
    public_id: str
    ip_address: str


class EnclaveConfig(TypedDict):
    base_url: str
    biolib_containers_pull_token: str
    compute_node_info: ComputeNodeInfo
    is_dev: bool



from biolib.biolib_logging import logger
from biolib.compute_node.job_worker.executors.types import SendSystemExceptionType


class ComputeProcessException(Exception):
    def __init__(self, original_error: Exception, biolib_error_code, send_system_exception: SendSystemExceptionType):
        super().__init__()

        send_system_exception(biolib_error_code)
        logger.error(original_error)



import os
import re

from biolib import cli_utils
from biolib.app.utils import run_job
from biolib.biolib_api_client import JobState
from biolib.biolib_api_client.app_types import App, AppVersionSlim
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.app.completed_process import CompletedProcess
from biolib.app.types import SemanticVersion, ParsedAppUri
from biolib.biolib_api_client.biolib_app_api import BiolibAppApi
from biolib.biolib_binary_format import ModuleInput
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger


class BioLibApp:

    def __init__(self, uri: str):
        parsed_uri = self._parse_uri(uri)

        self._app: App = BiolibAppApi.fetch_by_name(parsed_uri['account_handle'], parsed_uri['name'])

        if parsed_uri['version'] is None:
            self._selected_app_version: AppVersionSlim = self._app['active_version']
        else:
            parsed_version: SemanticVersion = parsed_uri['version']

            for version in self._app['app_versions']:
                if (
                        version['major'] == parsed_version['major'] and
                        version['minor'] == parsed_version['minor'] and
                        version['patch'] == parsed_version['patch']
                ):
                    self._selected_app_version = version
                    break

            if self._selected_app_version is None:
                raise Exception('Could not find application version referenced in URI')

        # TODO: Consider logging semantic app version here
        logger.info(f"Loaded package: {self._app['account_handle']}/{self._app['name']}")

    def __call__(self, args=None, stdin=None, files=None):
        module_input_serialized = self._get_serialized_module_input(args, stdin, files)

        job = BiolibJobApi.create(self._selected_app_version['public_id'])
        BiolibJobApi.update_state(job['public_id'], JobState.IN_PROGRESS.value)

        try:
            module_output = run_job(job, module_input_serialized)

            cli_utils.write_output_files(mapped_output_files=module_output['files'])

            try:
                BiolibJobApi.update_state(job_id=job['public_id'], state=JobState.COMPLETED.value)
            except Exception as error:  # pylint: disable=broad-except
                logger.warning(f'Could not update job state to completed:\n{error}')

            return CompletedProcess(
                exitcode=module_output['exit_code'],
                stderr=module_output['stderr'],
                stdout=module_output['stdout'],
            )

        except BioLibError as exception:
            logger.error(f'Compute failed with: {exception.message}')
            try:
                BiolibJobApi.update_state(job_id=job['public_id'], state=JobState.FAILED.value)
            except Exception as error:  # pylint: disable=broad-except
                logger.warning(f'Could not update job state to failed:\n{error}')

            raise exception

    @staticmethod
    def _get_serialized_module_input(args=None, stdin=None, files=None) -> bytes:
        if args is None:
            args = []

        if stdin is None:
            stdin = b''

        if isinstance(args, str):
            args = list(filter(lambda p: p != '', args.split(' ')))

        if not isinstance(args, list):
            raise Exception('The given input arguments must be list or str')

        if isinstance(stdin, str):
            stdin = stdin.encode('utf-8')

        if files is None:
            files = []
            for idx, arg in enumerate(args):
                if os.path.isfile(arg):
                    files.append(arg)
                    args[idx] = arg.split('/')[-1]

        cwd = os.getcwd()
        files_dict = {}

        for file in files:
            path = file
            if not file.startswith('/'):
                # make path absolute
                path = cwd + '/' + file

            arg_split = path.split('/')
            file = open(path, 'rb')
            path = '/' + arg_split[-1]

            files_dict[path] = file.read()
            file.close()

        module_input_serialized: bytes = ModuleInput().serialize(stdin=stdin, arguments=args, files=files_dict)
        return module_input_serialized

    @staticmethod
    def _parse_uri(uri: str) -> ParsedAppUri:
        uri_regex = r'((https:\/\/)?biolib\.com\/)?(?P<account>[\w-]+)\/(?P<name>[\w-]+)(\/version\/' \
                    r'(?P<version>(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)))?(\/)?'

        matches = re.search(uri_regex, uri)
        if matches is None:
            raise Exception('Application URI was incorrectly formatted. Please use the format: account_handle/app_name')

        return ParsedAppUri(
            account_handle=matches.group('account'),
            name=matches.group('name'),
            version=None if not matches.group('version') else SemanticVersion(
                major=int(matches.group('major')),
                minor=int(matches.group('minor')),
                patch=int(matches.group('patch')),
            )
        )



from biolib.biolib_api_client import RemoteHost
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger
from biolib.compute_node.remote_host_proxy import RemoteHostProxy
from biolib.typing_utils import List

_BIOLIB_ENCLAVE_REMOTE_HOSTS: List[RemoteHost] = [
    {'hostname': 'biolib-cloud-api.s3.amazonaws.com'},
    {'hostname': 'prod-eu-west-1-starport-layer-bucket.s3.eu-west-1.amazonaws.com'},
    {'hostname': 'biolib.com'},
    {'hostname': 'containers.biolib.com'},
    {'hostname': 'containers.staging.biolib.com'},
    {'hostname': 'staging-elb.biolib.com'},
]


def start_enclave_remote_hosts() -> None:
    logger.debug('Starting Docker network for enclave remote host proxies')
    docker = BiolibDockerClient.get_docker_client()
    public_network = docker.networks.create(
        driver='bridge',
        internal=False,
        name='biolib-enclave-remote-hosts-network',
    )

    biolib_remote_host_proxies: List[RemoteHostProxy] = []

    logger.debug('Starting enclave remote host proxies')
    for remote_host in _BIOLIB_ENCLAVE_REMOTE_HOSTS:
        remote_host_proxy = RemoteHostProxy(remote_host, public_network, internal_network=None, job_id=None)
        remote_host_proxy.start()
        biolib_remote_host_proxies.append(remote_host_proxy)

    logger.debug('Writing to enclave /etc/hosts')
    with open('/etc/hosts', mode='a') as hosts_file:
        for proxy in biolib_remote_host_proxies:
            ip_address = proxy.get_ip_address_on_network(public_network)
            hosts_file.write(f'\n{ip_address} {proxy.hostname}')



import http.server
import json
import socketserver
import threading
import os
import asyncio
import subprocess
import base64
import sys
import logging

from biolib.typing_utils import Optional
from biolib import utils
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger, TRACE
from biolib.compute_node.job_worker.executors.base_executor import BaseExecutor
from biolib.compute_node.job_worker.executors.types import LocalExecutorOptions
from biolib.pyppeteer.pyppeteer import launch, command  # type: ignore
from biolib.pyppeteer.pyppeteer.launcher import resolveExecutablePath, __chromium_revision__  # type: ignore

# specifically limit logs from pyppeteer
logging.getLogger('biolib.pyppeteer.pyppeteer').setLevel(logging.ERROR)
logging.getLogger('biolib.pyppeteer.pyppeteer.connection').setLevel(logging.ERROR)


class RequestHandler(http.server.SimpleHTTPRequestHandler):
    def log_message(self, format, *args):  # pylint: disable=redefined-builtin
        if logger.level == TRACE:
            http.server.SimpleHTTPRequestHandler.log_message(self, format, *args)


class PyppeteerExecutor(BaseExecutor):
    _chrome_executable_path: Optional[str] = None
    _no_sandbox: bool = True

    def execute_module(self, module_input_serialized: bytes) -> bytes:
        with socketserver.TCPServer(('127.0.0.1', 0), RequestHandler) as httpd:
            port = httpd.server_address[1]
            thread = threading.Thread(target=httpd.serve_forever)
            original_working_directory = os.getcwd()
            # TODO: figure out how we can avoid changing the current directory
            biolib_js_dir = os.path.join(utils.BIOLIB_PACKAGE_ROOT_DIR, 'biolib-js')
            os.chdir(biolib_js_dir)
            try:
                thread.start()
                module_output_serialized: bytes = asyncio.get_event_loop().run_until_complete(self._call_pyppeteer(
                    port,
                    self._options,
                    module_input_serialized,
                ))
                return module_output_serialized
            finally:
                os.chdir(original_working_directory)
                httpd.shutdown()
                thread.join()

    async def _call_pyppeteer(self, port: int, options: LocalExecutorOptions, module_input_serialized: bytes):
        if not self._chrome_executable_path:
            mac_chrome_path = '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome'
            if os.path.isfile(mac_chrome_path):
                self._chrome_executable_path = mac_chrome_path

        if not self._chrome_executable_path:
            linux_chrome_path = '/usr/lib/chromium-browser/chromium-browser'

            # special install for google colab
            if not os.path.isfile(linux_chrome_path) and 'google.colab' in sys.modules:
                subprocess.run('apt-get update', shell=True, check=True)
                subprocess.run('apt install chromium-chromedriver', shell=True, check=True)

            if os.path.isfile(linux_chrome_path):
                self._chrome_executable_path = linux_chrome_path

        resolved_path = resolveExecutablePath(None, __chromium_revision__)
        if not self._chrome_executable_path and resolved_path[1]:
            # if executable_path is not set explicit,
            # and resolveExecutablePath failed (== we got an error message back in resolved_path[1])
            logger.info('Installing dependencies...')
            os.environ['PYPPETEER_NO_PROGRESS_BAR'] = 'true'
            command.install()

        logger.info('Computing...')

        chrome_arguments = [
            '--disable-web-security',
        ]
        if self._no_sandbox:
            chrome_arguments.append('--no-sandbox')

        browser = await launch(args=chrome_arguments, executablePath=self._chrome_executable_path)

        # start new page
        page = await browser.newPage()

        await page.goto('http://localhost:' + str(port))

        def get_data():
            return base64.b64encode(module_input_serialized).decode('ascii')

        def set_progress_compute(value):
            logger.debug(f'Compute progress: {value}')

        def set_progress_initialization(value):
            logger.debug(f'Initialization progress: {value}')

        def add_log_message(value):
            logger.debug(f'Log message: {value}')

        await page.exposeFunction('getData', get_data)
        await page.exposeFunction('setProgressCompute', set_progress_compute)
        await page.exposeFunction('setProgressInitialization', set_progress_initialization)
        await page.exposeFunction('addLogMessage', add_log_message)

        api_client = BiolibApiClient.get()
        refresh_token_js_value = 'undefined' if not api_client.refresh_token else f'"{api_client.refresh_token}"'

        job = options['job'].copy()
        if 'custom_compute_node_url' in job:
            job.pop('custom_compute_node_url')

        job_json_string = json.dumps(job)

        output_serialized_js_bytes = await page.evaluate('''
        async function() {
          const refreshToken = ''' + refresh_token_js_value + ''';
          const baseUrl = \'''' + options['biolib_base_url'] + '''\';
          const job = ''' + job_json_string + ''';

          const { BioLibSingleton, AppClient } = window.BioLib;
          BioLibSingleton.setConfig({ baseUrl, refreshToken });
          AppClient.setApiClient(BioLibSingleton.get());

          const inputBase64 = await window.getData();
          const inputByteArray = Uint8Array.from(atob(inputBase64), c => c.charCodeAt(0));

          const jobUtils = {
              setProgressCompute: window.setProgressCompute,
              setProgressInitialization: window.setProgressInitialization,
              addLogMessage: window.addLogMessage,
          };

          try {
            const moduleOutput = await AppClient.runJob(job, inputByteArray, jobUtils);
            return moduleOutput.serialize();
          } catch(err) {
            return err.toString();
          }
        }
        ''')
        logger.debug('Closing browser')
        await browser.close()

        module_output_serialized = self._js_to_python_byte_array_converter(output_serialized_js_bytes)

        # TODO: check if module_output is correctly serialized
        if isinstance(module_output_serialized, bytes):
            return module_output_serialized
        else:
            raise BioLibError(module_output_serialized)

    @staticmethod
    def _js_to_python_byte_array_converter(js_encoded) -> bytes:
        try:
            return bytes(list([js_encoded[str(i)] for i in range(len(js_encoded))]))
        except Exception as error:
            logger.error('Failed to decode response from browser')
            logger.error(error)
            logger.error(js_encoded)
            raise BioLibError(js_encoded) from error



import os
import subprocess
import time
from typing import Dict, Optional

try:
    from typing_extensions import TypedDict
except ImportError:
    from typing import TypedDict

import requests
import docker.types  # type: ignore

from biolib.biolib_api_client import BiolibApiClient


class LargeFileSystemAttachResponse(TypedDict):
    aws_ebs_volume_id: str
    device_name: str


class DeviceInfo(TypedDict):
    attached_device_name: str
    aws_ebs_volume_id: str
    nvme_device_name: str


class LargeFileSystem:
    _EC2_INSTANCE_ID: Optional[str] = None

    def __init__(self, job_id: str, public_id: str, to_path: str):
        self.job_id = job_id
        self.public_id = public_id
        self.to_path = to_path

        self._mount_directory_path = f'/mnt/bl-lfs/{public_id}/'

    def get_as_docker_mount_object(self) -> docker.types.Mount:
        if not self._is_mounted():
            raise Exception('LargeFileSystem not mounted')

        return docker.types.Mount(
            read_only=True,
            source=self._mount_directory_path,
            target=self.to_path,
            type='bind',
        )

    def mount(self) -> None:
        if self._is_mounted():
            return

        try:
            api_client = BiolibApiClient.get()
            response = requests.post(
                url=f'{api_client.base_url}/lfs/{self.public_id}/attach/',
                json={
                    'ec2_instance_id': LargeFileSystem._get_ec2_instance_id(),
                    'job_id': self.job_id,
                },
            )
            if not response.ok:
                raise Exception(f'Failed to get response from {response.request.url}\n{response.content.decode()}')

            attach_response: LargeFileSystemAttachResponse = response.json()

            device_to_mount: Optional[DeviceInfo] = None
            for _ in range(10):
                time.sleep(0.5)
                device_to_mount = self._get_attached_ebs_devices().get(attach_response['aws_ebs_volume_id'])
                if device_to_mount is not None:
                    break

            if device_to_mount is None:
                raise Exception('LargeFileSystem device is not attached')

            subprocess.run(['sudo', 'mkdir', '--parents', self._mount_directory_path], check=True)
            subprocess.run(
                ['sudo', 'mount', f"/dev/{device_to_mount['nvme_device_name']}",
                 self._mount_directory_path],
                check=True,
            )
        except Exception as exception:
            raise Exception(f'Failed to mount LargeFileSystem with id {self.public_id}') from exception

    def unmount(self) -> None:
        if not self._is_mounted():
            return

        try:
            subprocess.run(['sudo', 'umount', '-d', self._mount_directory_path], check=True)
            # ensure directory is removed
            if os.path.exists(self._mount_directory_path):
                subprocess.run(['sudo', 'rm', '-r', self._mount_directory_path], check=True)
        except Exception as exception:
            raise Exception(f'Failed to unmount LargeFileSystem with id {self.public_id}') from exception

    def _is_mounted(self) -> bool:
        return os.path.exists(self._mount_directory_path)

    @staticmethod
    def _get_ec2_instance_id() -> str:
        if LargeFileSystem._EC2_INSTANCE_ID is None:
            # the IP address 169.254.169.254 is a link-local address and is valid only from the instance
            LargeFileSystem._EC2_INSTANCE_ID = requests.get('http://169.254.169.254/latest/meta-data/instance-id').text

        return LargeFileSystem._EC2_INSTANCE_ID

    @staticmethod
    def _get_attached_ebs_devices():
        lsblk_result = subprocess.run(['sudo', 'lsblk'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout_lines = lsblk_result.stdout.decode().splitlines()

        ebs_vol_id_to_device_info: Dict[str, DeviceInfo] = {}

        for line in stdout_lines:
            if line.startswith('nvme'):
                nvme_device_name = line.split()[0]
                ebsnvme_result = subprocess.run(
                    ['sudo', 'ebsnvme-id', f'/dev/{nvme_device_name}'],
                    check=False,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                # if command succeed the device is an EBS
                if ebsnvme_result.returncode == 0:
                    # we expect 2 lines
                    line1, line2 = ebsnvme_result.stdout.decode().splitlines()
                    device_info = DeviceInfo(
                        aws_ebs_volume_id=line1.replace('Volume ID: ', ''),
                        attached_device_name=line2,
                        nvme_device_name=nvme_device_name,
                    )
                    ebs_vol_id_to_device_info[device_info['aws_ebs_volume_id']] = device_info

        return ebs_vol_id_to_device_info



import os
from importlib_metadata import version, PackageNotFoundError

# try fetching version, if it fails (usually when in dev), add default
try:
    BIOLIB_PACKAGE_VERSION = version('pybiolib')
except PackageNotFoundError:
    BIOLIB_PACKAGE_VERSION = '0.0.0'

IS_DEV = os.getenv('BIOLIB_DEV', '').upper() == 'TRUE'

BIOLIB_PACKAGE_ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

BIOLIB_IS_RUNNING_IN_ENCLAVE = os.getenv('BIOLIB_IS_RUNNING_IN_ENCLAVE', '').upper() == 'TRUE'

BIOLIB_CLOUD_SKIP_PCR_VERIFICATION = os.getenv('BIOLIB_CLOUD_SKIP_PCR_VERIFICATION', '').upper() == 'TRUE'

RUN_DEV_JOB_ID = 'run-dev-mocked-job-id'



import io
import json
import logging
import socket
import shlex
import zipfile
from time import time
from queue import Queue
import multiprocessing
import os
import signal
from types import FrameType

import requests
from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from docker.models.networks import Network  # type: ignore

from biolib.compute_node.remote_host_proxy import RemoteHostProxy
from biolib.typing_utils import Optional, List, Dict
from biolib import utils
from biolib.biolib_api_client import ModuleEnvironment, Job, JobWrapper, Module, AppVersionOnJob
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger
from biolib.compute_node.job_worker.executors import PyppeteerExecutor, DockerExecutor, RemoteExecutor
from biolib.compute_node.job_worker.executors.base_executor import BaseExecutor
from biolib.compute_node.job_worker.executors.types import LocalExecutorOptions, StatusUpdate, RemoteExecuteOptions
from biolib.compute_node.socker_listener_thread import SocketListenerThread
from biolib.compute_node.socket_sender_thread import SocketSenderThread
from biolib.compute_node.job_worker.mappings import Mappings, path_without_first_folder
from biolib.compute_node.job_worker.utils import ComputeProcessException
from biolib.compute_node.utils import get_package_type, SystemExceptionCodes
from biolib.biolib_binary_format import AttestationDocument, SavedJob, SystemStatusUpdate, \
    RsaEncryptedAesPackage, AesEncryptedPackage, ModuleInput, SystemException

try:
    from biolib.compute_node.enclave.nitro_secure_module_utils import NitroSecureModuleUtils
except ImportError:
    pass

DEFAULT_BUFFER_SIZE = 1024
SOCKET_HOST = '127.0.0.1'


class JobWorkerProcess(multiprocessing.Process):

    # note: this method is run in the parent process
    def __init__(self, socket_port: int, log_level: int):
        super().__init__()
        self._socket_port = socket_port
        self._log_level = log_level

    # note: this method is run in the newly started process once called with .start()
    def run(self) -> None:
        _JobWorker(self._socket_port, self._log_level).run_handle_message_loop()


class _JobWorker:
    _STOP_HANDLE_MESSAGE_LOOP = b'STOP_HANDLE_MESSAGE_LOOP'

    def __init__(self, socket_port: int, log_level: int):
        try:
            logger.setLevel(log_level)

            # handle interrupt from keyboard (CTRL + C)
            signal.signal(signal.SIGINT, self._handle_exit_gracefully)
            # handle termination signal from parent
            signal.signal(signal.SIGTERM, self._handle_exit_gracefully)

            self._socket_port = socket_port
            self._received_messages_queue: Queue = Queue()
            self._messages_to_send_queue: Queue = Queue()

            self._app_version_id_to_runtime_zip: Dict[str, bytes] = {}
            self._jobs: Dict[str, Job] = {}
            self._root_job_wrapper: Optional[JobWrapper] = None

            self._remote_host_proxies: List[RemoteHostProxy] = []
            self._internal_network: Optional[Network] = None
            self._public_network: Optional[Network] = None

            if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
                self._nsm_util = NitroSecureModuleUtils()
                self._aes_key_buffer = b''
                logger.setLevel(logging.DEBUG)
        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_INIT_COMPUTE_PROCESS_VARIABLES.value,
                self._send_system_exception,
            ) from exception

        self._connect_to_parent()

    def _handle_exit_gracefully(self, signum: int, frame: FrameType) -> None:  # pylint: disable=unused-argument
        logger.debug(f'_JobWorker got exit signal {signal.Signals(signum).name}')  # pylint: disable=no-member
        self._received_messages_queue.put(self._STOP_HANDLE_MESSAGE_LOOP)
        self._cleanup()

    def run_handle_message_loop(self):
        logger.debug(f'Started JobWorkerProcess {os.getpid()}')
        while True:
            try:
                package = self._received_messages_queue.get()
                if package == self._STOP_HANDLE_MESSAGE_LOOP:
                    break

                package_type = get_package_type(package)
                if package_type == 'RsaEncryptedAesPackage':
                    encrypted_aes_key, iv, _, encrypted_data = RsaEncryptedAesPackage(package).deserialize()
                    self._aes_key_buffer = self._nsm_util.decrypt(encrypted_aes_key)
                    aes_key = AES.new(self._aes_key_buffer, AES.MODE_GCM, iv)

                    package = aes_key.decrypt(encrypted_data)
                    package_type = get_package_type(package)

                if package_type == 'SavedJob':
                    self._handle_save_job_wrapper(package)
                elif package_type == 'ModuleInput':
                    if not self._root_job_wrapper:
                        raise Exception('No job saved yet')

                    module_output_serialized = self._run_job(
                        job=self._root_job_wrapper['job'],
                        module_input_serialized=package,
                    )

                    if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
                        module_output_to_send = self._wrap_in_aes_encrypted_package(module_output_serialized)
                    else:
                        module_output_to_send = module_output_serialized

                    self._messages_to_send_queue.put(module_output_to_send)

                else:
                    logger.error(f'Package type from parent was not recognized: {package}')

                self._received_messages_queue.task_done()
            except ComputeProcessException:
                continue

            except Exception as exception:
                raise ComputeProcessException(
                    exception,
                    SystemExceptionCodes.UNKOWN_COMPUTE_PROCESS_ERROR.value,
                    self._send_system_exception,
                ) from exception

    def _cleanup(self) -> None:
        proxy_count = len(self._remote_host_proxies)
        if proxy_count > 0:
            proxy_cleanup_start_time = time()

            for proxy in self._remote_host_proxies:
                try:
                    proxy.terminate()
                except Exception as exception:  # pylint: disable=broad-except
                    logger.error(f'Failed to clean up remote host proxy: {exception}')

            self._remote_host_proxies = []
            logger.debug(f'Cleaned up {proxy_count} proxies in {time() - proxy_cleanup_start_time}')

        self._cleanup_network(self._internal_network)
        self._internal_network = None
        self._cleanup_network(self._public_network)
        self._public_network = None

    @staticmethod
    def _cleanup_network(network: Optional[Network]) -> None:
        if network:
            network_cleanup_start_time = time()
            network_name = network
            try:
                network.remove()
            except Exception as exception:  # pylint: disable=broad-except
                logger.error(f'Failed to clean up {network_name}: {exception}')
            logger.debug(f'Removed network {network_name} in {time() - network_cleanup_start_time}')

    def _handle_save_job_wrapper(self, package: bytes):
        job_wrapper_json_string = SavedJob(package).deserialize()
        job_wrapper: JobWrapper = json.loads(job_wrapper_json_string)
        self._root_job_wrapper = job_wrapper
        job = job_wrapper['job']
        self._jobs[job['public_id']] = job

        if job['app_version'].get('modules') is not None and BiolibDockerClient.is_docker_running():
            self._start_remote_host_proxies(job)

        # TODO: start downloading runtime zip already at this point

    def _start_remote_host_proxies(self, job: Job) -> None:
        job_id = job['public_id']
        remote_hosts = job['app_version']['remote_hosts']

        if len(remote_hosts) > 0:
            docker_client = BiolibDockerClient.get_docker_client()
            logger.debug(f'Creating networks for remote host proxies for job: {job_id}')
            try:
                self._internal_network = docker_client.networks.create(
                    name=f'biolib-sandboxed-network-{job_id}',
                    internal=True,
                    driver='bridge',
                )
                self._public_network = docker_client.networks.create(
                    name=f'biolib-proxy-network-{job_id}',
                    internal=False,
                    driver='bridge',
                )
            except Exception as exception:
                raise ComputeProcessException(
                    exception,
                    SystemExceptionCodes.FAILED_TO_CREATE_DOCKER_NETWORKS.value,
                    self._send_system_exception,
                ) from exception
            logger.debug(f'Starting remote host proxies for job: {job_id}')
            try:
                for remote_host in remote_hosts:
                    remote_host_proxy = RemoteHostProxy(
                        remote_host,
                        self._public_network,
                        self._internal_network,
                        job_id,
                    )
                    remote_host_proxy.start()
                    self._remote_host_proxies.append(remote_host_proxy)

            except Exception as exception:
                raise ComputeProcessException(
                    exception,
                    SystemExceptionCodes.FAILED_TO_START_REMOTE_HOST_PROXIES.value,
                    self._send_system_exception,
                ) from exception

            logger.debug(f'Completed startup of remote host proxies for job: {job_id}')

    def _run_app_version(self, app_version_id: str, module_input_serialized: bytes) -> bytes:
        job: Job = BiolibJobApi.create(app_version_id)
        self._jobs[job['public_id']] = job
        return self._run_job(job, module_input_serialized)

    def _run_job(self, job: Job, module_input_serialized: bytes) -> bytes:
        logger.info(f"Running job with id {job['public_id']}")
        if self._root_job_wrapper is None:
            raise Exception('root_job_wrapper was None')

        root_job = job
        while root_job['caller_job'] is not None and self._jobs.get(root_job['caller_job']) is not None:
            root_job = self._jobs[root_job['caller_job']]

        root_job_id = root_job['public_id']

        modules = job['app_version'].get('modules')

        if modules is None or not BiolibDockerClient.is_docker_running():
            return RemoteExecutor.execute_job(
                RemoteExecuteOptions(
                    biolib_base_url=self._root_job_wrapper['BASE_URL'],
                    job=job,
                    root_job_id=root_job_id,
                ),
                module_input_serialized,
            )

        main_module = self._get_module_from_name(modules, module_name='main')

        runtime_zip = self._get_runtime_zip_as_bytes(root_job_id=root_job_id, app_version=job['app_version'])

        module_output_serialized = self._run_module(
            LocalExecutorOptions(
                access_token=self._root_job_wrapper['access_token'],
                biolib_base_url=self._root_job_wrapper['BASE_URL'],
                biolib_containers_pull_token=self._root_job_wrapper.get('biolib_containers_pull_token'),
                job=job,
                module=main_module,
                root_job_id=root_job_id,
                runtime_zip_bytes=runtime_zip,
                send_status_update=self._send_status_update,
                send_system_exception=self._send_system_exception,
                internal_network=self._internal_network,
                remote_host_proxies=self._remote_host_proxies,
            ),
            module_input_serialized,
        )

        return module_output_serialized

    def _run_module(self, options: LocalExecutorOptions, module_input_serialized: bytes) -> bytes:
        module = options['module']
        executor_instance: BaseExecutor

        if module['environment'] == ModuleEnvironment.BIOLIB_APP.value:
            module_input = ModuleInput(module_input_serialized).deserialize()
            module_input_with_runtime_zip = self._add_runtime_zip_and_command_to_module_input(options, module_input)
            module_input_with_runtime_zip_serialized = ModuleInput().serialize(
                stdin=module_input_with_runtime_zip['stdin'],
                arguments=module_input_with_runtime_zip['arguments'],
                files=module_input_with_runtime_zip['files']
            )
            return self._run_app_version(module['image_uri'], module_input_with_runtime_zip_serialized)

        elif module['environment'] == ModuleEnvironment.BIOLIB_ECR.value:
            executor_instance = DockerExecutor(options)

        elif module['environment'] == ModuleEnvironment.BIOLIB_CUSTOM.value:
            executor_instance = PyppeteerExecutor(options)
        else:
            raise Exception(f"Unsupported module environment {module['environment']}")

        return executor_instance.execute_module(module_input_serialized)

    def _connect_to_parent(self):
        try:
            parent_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            parent_socket.connect((SOCKET_HOST, int(self._socket_port)))

        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_CONNECT_TO_WORKER_THREAD_SOCKET.value,
                self._send_system_exception,
            ) from exception

        try:
            SocketListenerThread(parent_socket, self._received_messages_queue).start()
            SocketSenderThread(parent_socket, self._messages_to_send_queue).start()
        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_START_SENDER_THREAD_OR_RECEIVER_THREAD.value,
                self._send_system_exception,
            ) from exception

        try:
            if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
                attestation_document = self._nsm_util.get_attestation_doc()
            else:
                attestation_document = b'Running locally'
        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_GET_ATTESTATION_DOCUMENT.value,
                self._send_system_exception,
            ) from exception

        self._messages_to_send_queue.put(AttestationDocument().serialize(attestation_document))

    # TODO: move this mapping logic to the ModuleInput class
    def _add_runtime_zip_and_command_to_module_input(self, options: LocalExecutorOptions, module_input):
        module = options['module']
        runtime_zip_byes = options['runtime_zip_bytes']
        # TODO: Figure out if we ever forward output mappings correctly (Do we only the mapping of the base image?)
        # TODO: Reuse much of the make_runtime_tar logic in BiolibDockerClient
        try:
            if runtime_zip_byes:
                runtime_zip = zipfile.ZipFile(io.BytesIO(runtime_zip_byes))
                source_mappings = Mappings(module['source_files_mappings'], module_input['arguments'])
                for zip_file_name in runtime_zip.namelist():
                    file_path = '/' + path_without_first_folder(zip_file_name)
                    mapped_file_names = source_mappings.get_mappings_for_path(file_path)
                    for mapped_file_name in mapped_file_names:
                        file_data = runtime_zip.read(zip_file_name)
                        module_input['files'].update({mapped_file_name: file_data})

            for command_part in reversed(shlex.split(module['command'])):
                module_input['arguments'].insert(0, command_part)

        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_CREATE_NEW_JOB.value,
                self._send_system_exception,
            ) from exception
        return module_input

    def _get_runtime_zip_as_bytes(self, root_job_id: str, app_version: AppVersionOnJob) -> Optional[bytes]:
        runtime_zip_url = app_version['client_side_executable_zip']

        # TODO: change this to a is None check when backend is fixed to not return empty string
        if not runtime_zip_url:
            return None

        runtime_zip_bytes: Optional[bytes] = self._app_version_id_to_runtime_zip.get(app_version['public_id'])

        if runtime_zip_bytes is None:
            if root_job_id == utils.RUN_DEV_JOB_ID:
                with open(runtime_zip_url, mode='rb') as runtime_zip_file:
                    runtime_zip_bytes = runtime_zip_file.read()

            else:
                self._send_status_update(StatusUpdate(progress=25, log_message='Downloading Source Files...'))

                start_time = time()
                logger.debug(f'Downloading runtime zip from S3 URL: {runtime_zip_url}')
                try:
                    runtime_zip_bytes = requests.get(runtime_zip_url).content
                except Exception as exception:
                    raise ComputeProcessException(
                        exception,
                        SystemExceptionCodes.FAILED_TO_DOWNLOAD_RUNTIME_ZIP.value,
                        self._send_system_exception,
                    ) from exception
                finally:
                    logger.debug(f'Downloading runtime zip took: {time() - start_time}s')

            self._app_version_id_to_runtime_zip[app_version['public_id']] = runtime_zip_bytes

        return runtime_zip_bytes

    @staticmethod
    def _get_module_from_name(modules: List[Module], module_name: str):
        for module in modules:
            if module['name'] == module_name:
                return module
        raise Exception(f'Could not find module with name {module_name}')

    def _wrap_in_aes_encrypted_package(self, package):
        iv = get_random_bytes(12)
        aes_key = AES.new(self._aes_key_buffer, AES.MODE_GCM, iv)
        encrypted_package, tag = aes_key.encrypt_and_digest(package)
        aes_encrypted_package = AesEncryptedPackage().serialize(iv, tag, encrypted_package)
        return aes_encrypted_package

    def _send_system_exception(self, biolib_exception_code: SystemExceptionCodes) -> None:
        system_exception_package = SystemException().serialize(biolib_exception_code)
        self._messages_to_send_queue.put(system_exception_package)

    def _send_status_update(self, status_update: StatusUpdate) -> None:
        try:
            status_update_package = SystemStatusUpdate().serialize(
                status_update['progress'],
                status_update['log_message'],
            )
            logger.debug(status_update['log_message'])
            self._messages_to_send_queue.put(status_update_package)
        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_SEND_STATUS_UPDATE.value,
                self._send_system_exception,
            ) from exception



import requests

from biolib.biolib_api_client.auth import BearerAuth
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger


class BiolibAppApi:
    @staticmethod
    def fetch_by_name(account_handle, app_name):
        response = requests.get(
            f'{BiolibApiClient.get().base_url}/api/apps/',
            params={'account_handle': account_handle, 'app_name': app_name},
            auth=BearerAuth(BiolibApiClient.get().access_token)
        )

        if not response.ok:
            raise Exception(response.content.decode())

        paginated_apps = response.json()
        if len(paginated_apps['results']) > 0:
            return paginated_apps['results'][0]
        else:
            raise BioLibError('App not found')

    @staticmethod
    def push_app_version(app_id, zip_binary, author, app_name, set_as_active):
        response = requests.post(
            f'{BiolibApiClient.get().base_url}/api/app_versions/',
            files={
                'app': (None, app_id),
                'set_as_active': (None, 'true' if set_as_active else 'false'),
                'state': (None, 'published'),
                'source_files_zip': zip_binary
            },
            auth=BearerAuth(BiolibApiClient.get().access_token)
        )
        if not response.ok:
            logger.error(f'Push failed for {author}/{app_name}:')
            raise BioLibError(response.text)

        # TODO: When response includes the version number, print the URL for the new app version
        logger.info(f'Successfully pushed app version for {author}/{app_name}.')
        return response.json()

    @staticmethod
    def update_app_version(app_version_id, data):
        response = requests.patch(
            f'{BiolibApiClient.get().base_url}/api/app_versions/{app_version_id}/',
            json=data,
            auth=BearerAuth(BiolibApiClient.get().access_token)
        )
        if not response.ok:
            logger.error(f'Failed to update app version {app_version_id}')
            raise BioLibError(response.text)

        return response.json()



from biolib.typing_utils import TypedDict, Optional

from biolib.biolib_api_client.common_types import SemanticVersion


class ParsedAppUri(TypedDict):
    name: str
    account_handle: str
    version: Optional[SemanticVersion]



import requests

from biolib.typing_utils import Optional
from biolib.biolib_api_client import RemoteHost
from biolib.compute_node.enclave.parent_api_types import VsockProxyResponse, EnclaveConfig


class ParentApi:
    _BASE_URL = 'http://127.0.0.1:5005'
    _enclave_config: Optional[EnclaveConfig] = None

    @staticmethod
    def get_enclave_config() -> EnclaveConfig:
        if ParentApi._enclave_config:
            return ParentApi._enclave_config

        response = requests.get(f'{ParentApi._BASE_URL}/config/', timeout=5)
        ParentApi._enclave_config = response.json()
        return ParentApi._enclave_config

    @staticmethod
    def deregister_and_shutdown() -> None:
        requests.post(url=f'{ParentApi._BASE_URL}/deregister_and_shutdown/', timeout=5)

    @staticmethod
    def start_vsock_proxy(remote_host: RemoteHost) -> VsockProxyResponse:
        response = requests.post(url=f'{ParentApi._BASE_URL}/vsock_proxy/', json=remote_host, timeout=5)
        vsock_proxy: VsockProxyResponse = response.json()
        return vsock_proxy

    @staticmethod
    def stop_vsock_proxy(vsock_proxy_id: str) -> None:
        requests.delete(url=f'{ParentApi._BASE_URL}/vsock_proxy/{vsock_proxy_id}/', timeout=5)

    @staticmethod
    def start_shutdown_timer(minutes_until_shutdown: int) -> None:
        requests.post(
            url=f'{ParentApi._BASE_URL}/start_shutdown_timer/',
            json={'minutes_until_shutdown': minutes_until_shutdown},
            timeout=5,
        )



import sys

# import and expose everything from the typing module
from typing import *  # pylint: disable=wildcard-import, unused-wildcard-import

if sys.version_info < (3, 8):
    from typing_extensions import TypedDict, Literal
